/*
 * Copyright (c) 2015, Freescale Semiconductor, Inc.
 * Copyright 2016-2017 NXP
 * All rights reserved.
 *
 * SPDX-License-Identifier: BSD-3-Clause
 */

/* FreeRTOS kernel includes. */
#include "FreeRTOS.h"
#include "task.h"
#include "queue.h"
#include "timers.h"
#include "semphr.h"

/* Freescale includes. */
#include "fsl_device_registers.h"
#include "fsl_debug_console.h"
#include "board.h"

#include "pin_mux.h"
#include "clock_config.h"

#include "fsl_rdc.h"
#include "fsl_gpio.h"
#include "fsl_gpc.h"
#include "fsl_gpt.h"
#include "fsl_iomuxc.h"
#include "fsl_mu.h"

/*******************************************************************************
 * Definitions
 ******************************************************************************/
#define USE_WAKEUP_TIMER        (1)
#define USE_WAKEUP_GPIO         (0)

#define APP_PowerUpSlot (5U)
#define APP_PowerDnSlot (6U)

#define RDC_DISABLE_A53_ACCESS 0xFC
#define RDC_DISABLE_M4_ACCESS 0xF3

#define TIME_DELAY_SLEEP      5000

/*
 * LPM state of M4 core
 */
typedef enum lpm_power_status_m4
{
    LPM_M4_STATE_RUN,
    LPM_M4_STATE_WAIT,
    LPM_M4_STATE_STOP,
} LPM_POWER_STATUS_M4;
/*
 * Clock Speed of M4 core
 */
typedef enum lpm_m4_clock_speed
{
    LPM_M4_HIGH_FREQ,
    LPM_M4_LOW_FREQ
} LPM_M4_CLOCK_SPEED;

extern void vPortSetupTimerInterrupt( void );

int32_t m4_chg_clk_flag = 0;
int32_t m4_vdelay_test = 0;

static LPM_POWER_STATUS_M4 m4_lpm_state = LPM_M4_STATE_RUN;

static SemaphoreHandle_t xDelayTestSemaphore = NULL;
static SemaphoreHandle_t xTimeoutSemaphore = NULL;

/* Task priorities. */
#define hello_task_PRIORITY      (configMAX_PRIORITIES - 2)
#define tickless_task_PRIORITY   (configMAX_PRIORITIES - 1)

/*******************************************************************************
 * Prototypes
 ******************************************************************************/
static void hello_task(void *pvParameters);
static void Tickless_task(void *pvParameters);

/*******************************************************************************
 * Code
 ******************************************************************************/
#if USE_WAKEUP_TIMER
void GPT1_IRQHandler(void)
{
    /* Clear interrupt flag.*/
    GPT_ClearStatusFlags(GPT1, kGPT_OutputCompare1Flag);
    GPT_StopTimer(GPT1);

    //PRINTF("GPT IRQ triggered\r\n");
/* Add for ARM errata 838869, affects Cortex-M4, Cortex-M4F, Cortex-M7, Cortex-M7F Store immediate overlapping
  exception return operation might vector to incorrect interrupt */
#if defined __CORTEX_M && (__CORTEX_M == 4U || __CORTEX_M == 7U)
    __DSB();
#endif
}
#endif

#if USE_WAKEUP_GPIO
void GPIO1_Combined_0_15_IRQHandler(void){
    //uint32_t iFlag;
    //iFlag = GPIO_PortGetInterruptFlags(GPIO1);
    //PRINTF("GPIO IRQ triggered 0x%x\r\n", iFlag);
    GPIO_PortClearInterruptFlags(GPIO1, 1U << 13);
#if defined __CORTEX_M && (__CORTEX_M == 4U || __CORTEX_M == 7U)
    __DSB();
#endif
}
#endif

void Peripheral_RdcSetting(void)
{
    rdc_periph_access_config_t periphConfig;

    RDC_GetDefaultPeriphAccessConfig(&periphConfig);
    /* Do not allow the A53 domain(domain0) to access the following peripherals. */
    periphConfig.policy = RDC_DISABLE_A53_ACCESS;
    periphConfig.periph = kRDC_Periph_UART4;
    RDC_SetPeriphAccessConfig(RDC, &periphConfig);
    //periphConfig.periph = kRDC_Periph_SNVS_HP;
    //RDC_SetPeriphAccessConfig(RDC, &periphConfig);
#if USE_WAKEUP_GPIO
    periphConfig.periph = kRDC_Periph_GPIO1;
    RDC_SetPeriphAccessConfig(RDC, &periphConfig);
#endif
#if USE_WAKEUP_TIMER
    periphConfig.periph = kRDC_Periph_GPT1;
    RDC_SetPeriphAccessConfig(RDC, &periphConfig);
#endif
    /* Do not allow the m4 domain(domain1) to access SAI3.
     * The purpose is to avoid system hang when A core to access SAI3 once M4 enters STOP mode.
     */
    //periphConfig.policy = RDC_DISABLE_M4_ACCESS;
    //periphConfig.periph = kRDC_Periph_SAI3;
    //RDC_SetPeriphAccessConfig(RDC, &periphConfig);
}

const char *LPM_MCORE_GetPowerStatusString(void)
{
    switch (m4_lpm_state)
    {
        case LPM_M4_STATE_RUN:
            return "RUN";
        case LPM_M4_STATE_WAIT:
            return "WAIT";
        case LPM_M4_STATE_STOP:
            return "STOP";
        default:
            return "UNKNOWN";
    }
}

void LPM_MCORE_ChangeM4Clock(LPM_M4_CLOCK_SPEED target)
{
    /* Change CCM Root to change M4 clock*/
    switch (target)
    {
        case LPM_M4_LOW_FREQ:
            if (CLOCK_GetRootMux(kCLOCK_RootM4) != kCLOCK_M4RootmuxOsc24M)
            {
                CLOCK_SetRootMux(kCLOCK_RootM4, kCLOCK_M4RootmuxOsc24M);
                CLOCK_SetRootDivider(kCLOCK_RootM4, 1U, 1U);
            }
            break;
        case LPM_M4_HIGH_FREQ:
            if (CLOCK_GetRootMux(kCLOCK_RootM4) != kCLOCK_M4RootmuxSysPll1)
            {
                CLOCK_SetRootDivider(kCLOCK_RootM4, 1U, 2U);
                CLOCK_SetRootMux(kCLOCK_RootM4, kCLOCK_M4RootmuxSysPll1); /* switch cortex-m4 to SYSTEM PLL1 */
            }
            break;
        default:
            break;
    }
}

void LPM_MCORE_SetPowerStatus(GPC_Type *base, LPM_POWER_STATUS_M4 targetPowerMode)
{
    gpc_lpm_config_t config;
    config.enCpuClk = false;
    config.enFastWakeUp = false;
    config.enDsmMask = false;
    config.enWfiMask = false;
    config.enVirtualPGCPowerdown = true;
    config.enVirtualPGCPowerup = true;
    switch (targetPowerMode)
    {
        case LPM_M4_STATE_RUN:
            GPC->LPCR_M4 = GPC->LPCR_M4 & (~GPC_LPCR_M4_LPM0_MASK);
            break;
        case LPM_M4_STATE_WAIT:
            GPC_EnterWaitMode(GPC, &config);
            break;
        case LPM_M4_STATE_STOP:
            GPC_EnterStopMode(GPC, &config);
            break;
        default:
            break;
    }

    m4_lpm_state = targetPowerMode;
}

void LPM_SysTickHandler( void )
{
	/* The SysTick runs at the lowest interrupt priority, so when this interrupt
	executes all interrupts must be unmasked.  There is therefore no need to
	save and then restore the interrupt mask value as its value is already
	known. */
	portDISABLE_INTERRUPTS();
	{
#if 1
            if (1 == m4_chg_clk_flag) {
                // Change to 24M
                LPM_MCORE_ChangeM4Clock(LPM_M4_LOW_FREQ);
                SystemCoreClockUpdate();
                vPortSetupTimerInterrupt();
                m4_chg_clk_flag = 0;
            } else if (2 == m4_chg_clk_flag) {
                // Change to 400M
                LPM_MCORE_ChangeM4Clock(LPM_M4_HIGH_FREQ);
                SystemCoreClockUpdate();
                vPortSetupTimerInterrupt();
                m4_chg_clk_flag = 0;
            }
#endif
		/* Increment the RTOS tick. */
		if( xTaskIncrementTick() != pdFALSE )
		{
			/* A context switch is required.  Context switching is performed in
			the PendSV interrupt.  Pend the PendSV interrupt. */
			portNVIC_INT_CTRL_REG = portNVIC_PENDSVSET_BIT;
		}
	}
	portENABLE_INTERRUPTS();
}

void PreSleepProcessing(void)
{
    DbgConsole_Deinit();
}

void PostSleepProcessing(void)
{
    DbgConsole_Init(BOARD_DEBUG_UART_INSTANCE, BOARD_DEBUG_UART_BAUDRATE, BOARD_DEBUG_UART_TYPE,
                    BOARD_DEBUG_UART_CLK_FREQ);
}

static uint8_t show_menu(void)
{
    PRINTF("\r\nSelect the desired operation \n\r\n");
    PRINTF("Press 1 for enter: WAIT     - Wait mode\r\n");
    PRINTF("Press 2 for enter: STOP     - Stop mode\r\n");
    PRINTF("Press 3 for wakeup A core by MU GIR INT\r\n");
    PRINTF("Press 4 for wakeup A core by MU TF INT\r\n");
    PRINTF("Press 5 for switch M4 clock to 24M\r\n");
    PRINTF("Press 6 for switch M4 clock back to 400M\r\n");
    PRINTF("Press 7 to test vTaskDelay for 20s\r\n");
    PRINTF("\r\nWaiting for power mode select..\r\n\r\n");

    /* Wait for user response */
    return GETCHAR();
}

#if USE_WAKEUP_TIMER
static void APP_InitWakeupGPT(void)
{
    gpt_config_t gptConfig;

    GPT_GetDefaultConfig(&gptConfig);
    gptConfig.enableRunInDoze = true;
    gptConfig.enableRunInStop = true;
    gptConfig.enableRunInWait = true;
    gptConfig.enableFreeRun = true;
    gptConfig.clockSource = kGPT_ClockSource_Osc;

    /* Initialize timer */
    GPT_Init(GPT1, &gptConfig);
    /* Divide GPT clock source frequency */
    GPT_SetOscClockDivider(GPT1, 2);
}

static void APP_SetWakeupConfig(uint8_t wakeup_sec)
{
    uint32_t gpt_clk_freq = 24000000 / (CLOCK_GetRootPreDivider(kCLOCK_RootGpt1)) / (CLOCK_GetRootPostDivider(kCLOCK_RootGpt1));

    /* Set timer period */
    gpt_clk_freq /= GPT_GetOscClockDivider(GPT1);
    GPT_SetOutputCompareValue(GPT1, kGPT_OutputCompare_Channel1, wakeup_sec * gpt_clk_freq - 1);
    /* Enable timer interrupt */
    GPT_EnableInterrupts(GPT1, kGPT_OutputCompare1InterruptEnable);

    /* Enable NVIC interrupt */
    EnableIRQ(GPT1_IRQn);
    /* Start counting */
    GPT_StartTimer(GPT1);
}

static uint8_t APP_GetWakeupTimeout(void)
{
    uint8_t timeout;

    while (1)
    {
        PRINTF("Select the wake up timeout in seconds.\r\n");
        PRINTF("The allowed range is 1s ~ 9s.\r\n");
        PRINTF("Enter 0 will keep in low power mode without wakeup.\r\n");
        PRINTF("Eg. enter 5 to wake up in 5 seconds.\r\n");
        PRINTF("\r\nWaiting for input timeout value...\r\n\r\n");

        timeout = GETCHAR();
        PRINTF("%c\r\n", timeout);
        if ((timeout >= '0') && (timeout <= '9'))
        {
            return timeout - '0';
        }
        PRINTF("Wrong value!\r\n");
    }
}

static uint8_t APP_GetWakeupConfig(void)
{
    uint8_t wakeupTimeout = 0;

    /* Wakeup source is LPIT, user should input wakeup timeout value. */
    wakeupTimeout = APP_GetWakeupTimeout();
    if (wakeupTimeout > 0)
    {
        PRINTF("Will wakeup in %d seconds.\r\n", wakeupTimeout);
    }
    else
    {
        PRINTF("Will stay without wakeup forever.\r\n");
    }

    return wakeupTimeout;
}
#endif

/*!
 * @brief Main function
 */
int main(void)
{
    int32_t i = 0;

    /* Init board hardware. */
    /* Board specific RDC settings */
    BOARD_RdcInit();
    Peripheral_RdcSetting();
    BOARD_InitPins();
    BOARD_BootClockRUN();
    BOARD_InitDebugConsole();
    BOARD_InitMemory();

    MU_Init(MUB);

    /*
     * In order to wakeup M4 from LPM, all PLLCTRLs need to be set to "NeededRun"
     */
    for (i = 0; i < 39; i++)
    {
        CCM->PLL_CTRL[i].PLL_CTRL = kCLOCK_ClockNeededRun;
    }

    /* Configure GPC */
    GPC_Init(BOARD_GPC_BASEADDR, APP_PowerUpSlot, APP_PowerDnSlot);
    GPC_EnableIRQ(BOARD_GPC_BASEADDR, SNVS_IRQn);

#if USE_WAKEUP_TIMER
    CLOCK_SetRootMux(kCLOCK_RootGpt1, kCLOCK_GptRootmuxOsc24M);      /* Set GPT source to Osc24 MHZ */
    CLOCK_SetRootDivider(kCLOCK_RootGpt1, 1U, 1U);
    APP_InitWakeupGPT();
    GPC_EnableIRQ(BOARD_GPC_BASEADDR, GPT1_IRQn);
#endif
#if USE_WAKEUP_GPIO
    gpio_pin_config_t gpioConfig = {kGPIO_DigitalInput, 0, kGPIO_IntHighLevel};
    IOMUXC_SetPinMux(IOMUXC_GPIO1_IO13_GPIO1_IO13, 0U);
    IOMUXC_SetPinConfig(IOMUXC_GPIO1_IO13_GPIO1_IO13, IOMUXC_SW_PAD_CTL_PAD_DSE(6U) | IOMUXC_SW_PAD_CTL_PAD_FSEL(2U) | IOMUXC_SW_PAD_CTL_PAD_ODE_MASK | IOMUXC_SW_PAD_CTL_PAD_HYS_MASK); 
    GPIO_PinInit(GPIO1, 13, &gpioConfig);
    GPIO_PortClearInterruptFlags(GPIO1, 1U << 13);
    GPIO_PortEnableInterrupts(GPIO1, 1U << 13);
    EnableIRQ(GPIO1_Combined_0_15_IRQn);
    GPC_EnableIRQ(BOARD_GPC_BASEADDR, GPIO1_Combined_0_15_IRQn);
#endif
    PRINTF("\r\n####################  LOW POWER DEMO ####################\n\r\n");
    PRINTF("    Build Time: %s--%s \r\n", __DATE__, __TIME__);

    xDelayTestSemaphore = xSemaphoreCreateBinary();
    xTimeoutSemaphore = xSemaphoreCreateBinary();

    /*Create tickless task*/
    if (xTaskCreate(Tickless_task, "Tickless_task", configMINIMAL_STACK_SIZE + 100, NULL, tickless_task_PRIORITY,
                    NULL) != pdPASS)
    {
        PRINTF("Task creation failed!.\r\n");
        while (1)
            ;
    }

    if (xTaskCreate(hello_task, "Hello_task", configMINIMAL_STACK_SIZE + 100, NULL, hello_task_PRIORITY, NULL) !=
        pdPASS)
    {
        PRINTF("Task creation failed!.\r\n");
        while (1)
            ;
    }
    vTaskStartScheduler();
    for (;;)
        ;
}

/* Tickless Task */
static void Tickless_task(void *pvParameters)
{
    static TickType_t init_tick_cnt = 0;

    for (;;)
    {
        if (xSemaphoreTake(xDelayTestSemaphore, portMAX_DELAY) == pdTRUE)
        {
            PRINTF("Start vTaskDelay Test\r\n");
            init_tick_cnt = xTaskGetTickCount();
            for (;;)
            {
                if (m4_vdelay_test) {
                    TickType_t tick_cnt = xTaskGetTickCount();
                    PRINTF("%d\r\n", tick_cnt - init_tick_cnt);
                    vTaskDelay(TIME_DELAY_SLEEP);
                    if ((tick_cnt - init_tick_cnt) > 4 * TIME_DELAY_SLEEP) {
                        m4_vdelay_test = 0;
                    }
                } else {
                    xSemaphoreGive(xTimeoutSemaphore);
                    break;
                }
            }
        }
    }
}

extern uint32_t SystemCoreClock;
/*!
 * @brief Task responsible for printing of "Hello world." message.
 */
static void hello_task(void *pvParameters)
{
    while (1)
    {
        uint8_t ch_input;

        ch_input = show_menu();
        PRINTF("Input: %c\r\n", ch_input);
        switch (ch_input) {
        case '1':
            {
                // Wait
#if USE_WAKEUP_TIMER
                uint8_t wakeup_timeout = 0;

                wakeup_timeout = APP_GetWakeupConfig();
                if (wakeup_timeout > 0)
                {
                    APP_SetWakeupConfig(wakeup_timeout);
                }
#endif

                LPM_MCORE_ChangeM4Clock(LPM_M4_LOW_FREQ);
                LPM_MCORE_SetPowerStatus(BOARD_GPC_BASEADDR, LPM_M4_STATE_WAIT);
                PRINTF("\r\nMode:%s\r\n", LPM_MCORE_GetPowerStatusString());
                PreSleepProcessing();
                __DSB();
                __ISB();
                __WFI();
                PostSleepProcessing();
                LPM_MCORE_ChangeM4Clock(LPM_M4_HIGH_FREQ);
                LPM_MCORE_SetPowerStatus(BOARD_GPC_BASEADDR, LPM_M4_STATE_RUN);
                PRINTF("\r\nMode:%s\r\n", LPM_MCORE_GetPowerStatusString());
            }
            break;
        case '2':
            {
                // Stop
#if USE_WAKEUP_TIMER
                uint8_t wakeup_timeout = 0;

                wakeup_timeout = APP_GetWakeupConfig();
                if (wakeup_timeout > 0)
                {
                    APP_SetWakeupConfig(wakeup_timeout);
                }
#endif

                LPM_MCORE_ChangeM4Clock(LPM_M4_LOW_FREQ);
                LPM_MCORE_SetPowerStatus(BOARD_GPC_BASEADDR, LPM_M4_STATE_STOP);
                PRINTF("\r\nMode:%s\r\n", LPM_MCORE_GetPowerStatusString());
                PreSleepProcessing();
                __DSB();
                __ISB();
                __WFI();
                PostSleepProcessing();
                LPM_MCORE_ChangeM4Clock(LPM_M4_HIGH_FREQ);
                LPM_MCORE_SetPowerStatus(BOARD_GPC_BASEADDR, LPM_M4_STATE_RUN);
                PRINTF("\r\nMode:%s\r\n", LPM_MCORE_GetPowerStatusString());
            }
            break;
        case '3':
            {
                // Wakeup by GIR
                MU_TriggerInterrupts(MUB, kMU_GenInt0InterruptTrigger);
                PRINTF("Set mu interrupt kMU_GenInt0InterruptTrigger\r\n");
            }
            break;
        case '4':
            {
                // Wakeup by TF
                MU_SendMsg(MUB, 1, 2);
        	PRINTF("set mu interrupt MU_SendMsg\r\n");
            }
            break;
        case '5':
            // Switch M4 clock to 24M
            // Clock will be updated in xPortSysTickHandler
            m4_chg_clk_flag = 1;
            while (m4_chg_clk_flag)
                ;
            PRINTF("Clock changed to %d\r\n", SystemCoreClock);
            break;
        case '6':
            // Switch M4 clock back to 400M
            // Clock will be updated in xPortSysTickHandler
            m4_chg_clk_flag = 2;
            while (m4_chg_clk_flag)
                ;
            PRINTF("Clock changed to %d\r\n", SystemCoreClock);
            break;
        case '7':
            // Test vDelay
            {
                TaskHandle_t h_tickless_task;
                /*Create tickless task*/
                if (xTaskCreate(Tickless_task, "Tickless_task", configMINIMAL_STACK_SIZE + 100, NULL, tickless_task_PRIORITY,
                                &h_tickless_task) != pdPASS)
                {
                    PRINTF("Task creation failed!.\r\n");
                    while (1)
                        ;
                }
                PRINTF("\r\nTick count :\r\n");
                m4_vdelay_test = 1;
                xSemaphoreGive(xDelayTestSemaphore);
                if (xSemaphoreTake(xTimeoutSemaphore, portMAX_DELAY) == pdTRUE)
                {
                    vTaskDelete(h_tickless_task);
                }
            }
            break;
        default:
            break;
        }
    }
}
